# validated: 2018-11-27 DS 8eafe7f32561 cpp/ntcore_cpp.cpp

from .connection_notifier import ConnectionNotifier
from .dispatcher import Dispatcher
from .ds_client import DsClient
from .entry_notifier import EntryNotifier
from .rpc_server import RpcServer
from .storage import Storage

from ntcore.constants import NT_NOTIFY_IMMEDIATE, NT_NOTIFY_NEW

_is_new = NT_NOTIFY_IMMEDIATE | NT_NOTIFY_NEW


class NtCoreApi(object):
    """
        Internal NetworkTables API wrapper
        
        In theory you could create multiple instances of this
        and talk to multiple NT servers or create multiple
        NT servers... though, I don't really know why one
        would want to do this.
    """

    def __init__(self, entry_creator, verbose=False):
        self.conn_notifier = ConnectionNotifier(verbose=verbose)
        self.entry_notifier = EntryNotifier(verbose=verbose)
        self.rpc_server = RpcServer(verbose=verbose)
        self.storage = Storage(self.entry_notifier, self.rpc_server, entry_creator)
        self.dispatcher = Dispatcher(self.storage, self.conn_notifier, verbose=verbose)
        self.ds_client = DsClient(self.dispatcher, verbose=verbose)

        self._init_table_functions()

    def stop(self):
        self.ds_client.stop()
        self.dispatcher.stop()
        self.rpc_server.stop()
        self.entry_notifier.stop()
        self.conn_notifier.stop()
        self.storage.stop()

    def destroy(self):
        self.ds_client = None
        self.dispatcher = None
        self.rpc_server = None
        self.entry_notifier = None
        self.entry_notifier = None
        self.conn_notifier = None
        self.storage = None

    #
    # Table functions (inline because they're called often)
    #

    def _init_table_functions(self):
        self.getEntry = self.storage.getEntry
        self.getEntryId = self.storage.getEntryId
        self.getEntries = self.storage.getEntries
        self.getEntryNameById = self.storage.getEntryNameById
        self.getEntryTypeById = self.storage.getEntryTypeById
        self.getEntryValue = self.storage.getEntryValue
        self.setDefaultEntryValue = self.storage.setDefaultEntryValue
        self.setDefaultEntryValueById = self.storage.setDefaultEntryValueById
        self.setEntryValue = self.storage.setEntryValue
        self.setEntryValueById = self.storage.setEntryValueById
        self.setEntryTypeValue = self.storage.setEntryTypeValue
        self.setEntryTypeValueById = self.storage.setEntryTypeValueById
        self.setEntryFlags = self.storage.setEntryFlags
        self.setEntryFlagsById = self.storage.setEntryFlagsById
        self.getEntryFlags = self.storage.getEntryFlags
        self.getEntryFlagsById = self.storage.getEntryFlagsById
        self.deleteEntry = self.storage.deleteEntry
        self.deleteEntryById = self.storage.deleteEntryById
        self.deleteAllEntries = self.storage.deleteAllEntries
        self.getEntryInfo = self.storage.getEntryInfo
        self.getEntryInfoById = self.storage.getEntryInfoById

    #
    # Entry notification
    #

    def addEntryListener(self, prefix, callback, flags):
        return self.storage.addListener(prefix, callback, flags)

    def addEntryListenerById(self, local_id, callback, flags):
        return self.storage.addListenerById(local_id, callback, flags)

    def addEntryListenerByIdEx(
        self, fromobj, key, local_id, callback, flags, paramIsNew
    ):
        if paramIsNew:

            def listener(item):
                key_, value_, flags_, _ = item
                callback(fromobj, key, value_.value, (flags_ & _is_new) != 0)

        else:

            def listener(item):
                key_, value_, flags_, _ = item
                callback(fromobj, key, value_.value, flags_)

        return self.storage.addListenerById(local_id, listener, flags)

    def createEntryListenerPoller(self):
        return self.entry_notifier.createPoller()

    def destroyEntryListenerPoller(self, poller_uid):
        self.entry_notifier.removePoller(poller_uid)

    def addPolledEntryListener(self, poller_uid, prefix, flags):
        return self.storage.addPolledListener(poller_uid, prefix, flags)

    def addPolledEntryListenerById(self, poller_uid, local_id, flags):
        return self.storage.addPolledListenerById(poller_uid, local_id, flags)

    def pollEntryListener(self, poller_uid, timeout=None):
        return self.entry_notifier.poll(poller_uid, timeout=timeout)

    def cancelPollEntryListener(self, poller_uid):
        self.entry_notifier.cancelPoll(poller_uid)

    def removeEntryListener(self, listener_uid):
        self.entry_notifier.remove(listener_uid)

    def waitForEntryListenerQueue(self, timeout):
        return self.entry_notifier.waitForQueue(timeout)

    #
    # Connection notifications
    #

    def addConnectionListener(self, callback, immediate_notify):
        return self.dispatcher.addListener(callback, immediate_notify)

    def createConnectionListenerPoller(self):
        return self.conn_notifier.createPoller()

    def destroyConnectionListenerPoller(self, poller_uid):
        self.conn_notifier.removePoller(poller_uid)

    def addPolledConnectionListener(self, poller_uid, immediate_notify):
        return self.dispatcher.addPolledListener(poller_uid, immediate_notify)

    def pollConnectionListener(self, poller_uid, timeout=None):
        return self.conn_notifier.poll(poller_uid, timeout=timeout)

    def cancelPollConnectionListener(self, poller_uid):
        self.conn_notifier.cancelPoll(poller_uid)

    def removeConnectionListener(self, listener_uid):
        self.conn_notifier.remove(listener_uid)

    def waitForConnectionListenerQueue(self, timeout):
        return self.conn_notifier.waitForQueue(timeout)

    #
    # TODO: RPC stuff not currently implemented
    #       .. there's probably a good pythonic way to implement
    #          it, but I don't really want to deal with it now.
    #          If you care, submit a PR.
    #
    #          I would have the caller register the server function
    #          via a docstring.
    #

    #
    # Client/Server Functions
    #

    def setNetworkIdentity(self, name):
        self.dispatcher.setIdentity(name)

    def getNetworkMode(self):
        return self.dispatcher.getNetworkMode()

    # python-specific
    def startTestMode(self, is_server):
        if self.dispatcher.startTestMode(is_server):
            self.storage.m_server = is_server
            return True
        else:
            return False

    def startServer(self, persist_filename, listen_address, port):
        return self.dispatcher.startServer(persist_filename, listen_address, port)

    def stopServer(self):
        self.dispatcher.stop()

    def startClient(self):
        return self.dispatcher.startClient()

    def stopClient(self):
        self.dispatcher.stop()

    def setServer(self, server_or_servers):
        self.dispatcher.setServer(server_or_servers)

    def setServerTeam(self, teamNumber, port):
        self.dispatcher.setServerTeam(teamNumber, port)

    def startDSClient(self, port):
        self.ds_client.start(port)

    def stopDSClient(self):
        self.ds_client.stop()

    def setUpdateRate(self, interval):
        self.dispatcher.setUpdateRate(interval)

    def flush(self):
        self.dispatcher.flush()

    def getRemoteAddress(self):
        if not self.dispatcher.isServer():
            for conn in self.dispatcher.getConnections():
                return conn.remote_ip

    def getIsConnected(self):
        return self.dispatcher.isConnected()

    def setVerboseLogging(self, verbose):
        self.conn_notifier.setVerboseLogging(verbose)
        self.dispatcher.setVerboseLogging(verbose)
        self.entry_notifier.setVerboseLogging(verbose)
        self.rpc_server.setVerboseLogging(verbose)

    #
    # Persistence
    #

    def savePersistent(self, filename):
        return self.storage.savePersistent(filename, periodic=False)

    def loadPersistent(self, filename):
        return self.storage.loadPersistent(filename)

    def saveEntries(self, filename, prefix):
        return self.storage.saveEntries(prefix, filename=filename)

    def loadEntries(self, filename, prefix):
        return self.storage.loadEntries(filename=filename, prefix=prefix)



# notrack
"""
    Instrumentation for finding deadlocks in networktables
"""

from __future__ import print_function

import inspect
import socket
import threading
import time

# Number of seconds to block
sock_block_period = None

# List of locks that can be acquired from the main thread
main_locks = ["entry_lock", "trans_lock"]

# List of locks that are allowed to be held when accessing a socket
# -> must never be locks that can be acquired by the main thread
sock_locks = ["client_conn_lock", "server_conn_lock", "write_lock"]

# Dictionary of locks
# key: name, value: locks that can be held when acquiring the lock
locks = {
    # Never held by robot thread
    "client_conn_lock": ["client_conn_lock"],
    "entry_lock": ["entry_lock", "client_conn_lock"],
    # Never held by robot thread
    "server_conn_lock": ["server_conn_lock"],
    "trans_lock": [
        "entry_lock",
        # Not 100% sure if this should be allowed
        # -> this only happens when NetworkTable API calls are made from
        #    a fired listener
        "server_conn_lock",
        "client_conn_lock",
    ],
    # Never held by robot thread
    "write_lock": ["client_conn_lock", "server_conn_lock", "write_lock"],
}

local = threading.local()


class WrappedLock(threading._PyRLock):
    def __init__(self, name):
        threading._PyRLock.__init__(self)
        self._name = name
        self._nt_creator = _get_caller()

    def acquire(self, blocking=True, timeout=-1):

        # This check isn't strictly true..
        if isinstance(threading.current_thread(), threading._MainThread):
            assert self._name in main_locks, (
                "%s cannot be held in main thread" % self._name
            )

        if not hasattr(local, "held_locks"):
            local.held_locks = []

        for lock in local.held_locks:
            assert (
                lock in locks[self._name]
            ), "Cannot hold %s when trying to acquire %s" % (lock._name, self._name)

        retval = threading._PyRLock.acquire(self, blocking=blocking, timeout=timeout)
        if retval != False:
            local.held_locks.append(self)

    __enter__ = acquire

    def release(self):
        threading._PyRLock.release(self)
        assert local.held_locks[-1] == self
        local.held_locks.pop()

    # Allow this to be used in comparisons

    def __eq__(self, other):
        if isinstance(other, str):
            return self._name.__eq__(other)
        else:
            return self._name.__eq__(other._name)

    def __cmp__(self, other):
        if isinstance(other, str):
            return self._name.__cmp__(other)
        else:
            return self._name.__cmp__(other._name)

    def __hash__(self):
        return self._name.__hash__()


def create_tracked_rlock(name):
    assert name in locks
    return WrappedLock(name)


def assert_not_locked(t):

    assert not isinstance(
        threading.current_thread(), threading._MainThread
    ), "Should not make socket calls from main thread"

    if not hasattr(local, "held_locks"):
        local.held_locks = []

    for lock in local.held_locks:
        assert lock in sock_locks, "ERROR: network %s was made while holding %s" % (
            t,
            lock._name,
        )


class WrappedFile:
    def __init__(self, file):
        self._file = file

    def write(self, data):
        assert_not_locked("write")
        if sock_block_period:
            time.sleep(sock_block_period)
        return self._file.write(data)

    def read(self, *args, **kwargs):
        assert_not_locked("read")
        if sock_block_period:
            time.sleep(sock_block_period)
        return self._file.read(*args, **kwargs)

    def __getattr__(self, attr):
        return getattr(self._file, attr)


def blocking_sock_makefile(s, mode):
    return WrappedFile(s.makefile(mode))


def blocking_sock_create_connection(address):
    assert_not_locked("connect")
    if sock_block_period:
        time.sleep(sock_block_period)
    return socket.create_connection(address)


def _get_caller():
    curframe = inspect.currentframe()
    calframe = inspect.getouterframes(curframe, 3)
    return "%s:%s %s" % (calframe[3][1], calframe[3][2], calframe[3][3])



from __future__ import print_function

from os.path import dirname, exists, join
import sys, subprocess

from setuptools import find_packages, setup

setup_dir = dirname(__file__)
git_dir = join(setup_dir, ".git")
version_file = join(setup_dir, "ntcore", "version.py")

# Automatically generate a version.py based on the git version
if exists(git_dir):
    p = subprocess.Popen(
        ["git", "describe", "--tags", "--long", "--dirty=-dirty"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out, err = p.communicate()
    # Make sure the git version has at least one tag
    if err:
        print("Error: You need to create a tag for this repo to use the builder")
        sys.exit(1)

    # Convert git version to PEP440 compliant version
    # - Older versions of pip choke on local identifiers, so we can't include the git commit
    v, commits, local = out.decode("utf-8").rstrip().split("-", 2)
    if commits != "0" or "-dirty" in local:
        v = "%s.post0.dev%s" % (v, commits)

    # Create the version.py file
    with open(version_file, "w") as fp:
        fp.write("# Autogenerated by setup.py\n__version__ = '{0}'".format(v))

with open(version_file, "r") as fp:
    exec(fp.read(), globals())

with open(join(setup_dir, "README.rst"), "r") as readme_file:
    long_description = readme_file.read()

setup(
    name="pynetworktables",
    version=__version__,
    description="A pure Python implementation of NetworkTables, used for robot communications in the FIRST Robotics Competition.",
    long_description=long_description,
    author="Dustin Spicuzza, Peter Johnson",
    author_email="robotpy@googlegroups.com",
    url="https://github.com/robotpy/pynetworktables",
    keywords="frc first robotics wpilib networktables",
    packages=find_packages(exclude="tests"),
    python_requires=">=3.5",
    license="BSD-3-Clause",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering",
    ],
)
